# -*- coding: utf-8 -*-

from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.cache import invalidate_cachekey_volatile_for
from plone.api.exc import InvalidParameterError
from plone.memoize import ram
from Products.PlonePAS.plugins.role import GroupAwareRoleManager
from Products.PluggableAuthService.plugins.ZODBRoleManager import ZODBRoleManager
from smtplib import SMTP_SSL
from zope.sendmail.mailer import SMTPMailer

import pkg_resources


try:
    pkg_resources.get_distribution("collective.solr")
except pkg_resources.DistributionNotFound:
    HAS_SOLR = False
else:
    from collective.solr.indexer import SolrIndexProcessor
    HAS_SOLR = True


def solr_index(self, obj, attributes=None):
    # Fix issue https://github.com/collective/collective.solr/issues/189
    if attributes is not None:
        attributes = None
    return SolrIndexProcessor._old_index(self, obj, attributes)


def ssl_makeMailer(self):
    """ Create a SMTPMailer with SMTP_SSL if port is 465 """
    mailer = SMTPMailer(hostname=self.smtp_host, port=int(self.smtp_port), username=self.smtp_uid or None,
                        password=self.smtp_pwd or None, force_tls=self.force_tls)
    if self.smtp_port in (465, '465'):
        mailer.smtp = SMTP_SSL
    return mailer


def assignRolesToPrincipal(self, roles, principal_id, REQUEST=None):  # noqa
    GroupAwareRoleManager._old_assignRolesToPrincipal(self, roles, principal_id, REQUEST)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')


def assignRoleToPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_assignRoleToPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')


def removeRoleFromPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_removeRoleFromPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for('_users_groups_value')


def _listAllowedRolesAndUsers_cachekey(method, self, user):
    '''cachekey method for self._listAllowedRolesAndUsers.'''
    date = get_cachekey_volatile('_users_groups_value')
    return date, user.getId()


@ram.cache(_listAllowedRolesAndUsers_cachekey)
def _listAllowedRolesAndUsers(self, user):
    """Monkeypatch to use get_plone_groups_for_user instead getGroups.
       Moreover store this in the REQUEST."""
    # Makes sure the list includes the user's groups.
    result = user.getRoles()
    if 'Anonymous' in result:
        # The anonymous user has no further roles
        return ['Anonymous']
    result = list(result)
    # XXX change, replaced getGroups by get_plone_groups_for_user
    # if hasattr(aq_base(user), 'getGroups'):
    #     groups = ['user:%s' % x for x in user.getGroups()]
    try:
        groups = get_plone_groups_for_user(user=user)
    except InvalidParameterError:
        groups = user.getGroups()
    if groups:
        groups = ['user:%s' % x for x in groups]
        result = result + groups
    # Order the arguments from small to large sets
    result.insert(0, 'user:%s' % user.getId())
    result.append('Anonymous')
    return result



# -*- coding: utf-8 -*-

from datetime import datetime
from imio.helpers.cache import _generate_params_key
from imio.helpers.cache import cleanForeverCache
from imio.helpers.cache import cleanRamCache
from imio.helpers.cache import cleanRamCacheFor
from imio.helpers.cache import cleanVocabularyCacheFor
from imio.helpers.cache import generate_key
from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import invalidate_cachekey_volatile_for
from imio.helpers.cache import obj_modified
from imio.helpers.cache import VOLATILE_ATTR
from imio.helpers.cache import volatile_cache_with_parameters
from imio.helpers.cache import volatile_cache_without_parameters
from imio.helpers.testing import IntegrationTestCase
from persistent.mapping import PersistentMapping
from plone import api
from plone.app.testing import login
from plone.app.testing import logout
from plone.memoize import forever
from plone.memoize import ram
from plone.memoize.instance import Memojito
from plone.memoize.interfaces import ICacheChooser
from Products.CMFCore.utils import _getAuthenticatedUser
from Products.PlonePAS.plugins.ufactory import PloneUser
from zope.annotation import IAnnotations
from zope.component import getUtility
from zope.component import queryUtility
from zope.ramcache.interfaces.ram import IRAMCache
from zope.schema.interfaces import IVocabularyFactory

import os
import time
import transaction


memPropName = Memojito.propname


def ramCachedMethod_cachekey(method, portal, param, pass_volatile_method=False):
    '''cachekey method for ramCachedMethod.'''
    # used for test_get_cachekey_volatile_method_cleanup
    if pass_volatile_method:
        date = get_cachekey_volatile('test_cache.ramCachedMethod', method)
    else:
        date = get_cachekey_volatile('test_cache.ramCachedMethod')
    return date, param


@ram.cache(ramCachedMethod_cachekey)
def ramCachedMethod(portal, param, pass_volatile_method=False):
    """ """
    return portal.REQUEST.get('ramcached', None)


@volatile_cache_without_parameters
def volatile_without_parameters_cached(portal):
    return portal.REQUEST.get('volatile_without_parameters_cached', None)


@volatile_cache_with_parameters
def volatile_with_parameters_cached(portal, param):
    return portal.REQUEST.get('volatile_with_parameters_cached', None)


@forever.memoize
def forever_cached_date():
    return datetime.now()


class TestCacheModule(IntegrationTestCase):
    """
    Test all helper methods of cache module.
    """

    def test_cleanVocabularyCacheFor(self):
        """
        This helper method cleans instance.memoize cache defined on a vocabulary.
        """
        self.request.set('vocab_values', ('1', '2'))
        vocab = queryUtility(IVocabularyFactory,
                             "imio.helpers.testing.testingvocabulary")
        # not cached for now
        self.assertFalse(getattr(vocab, memPropName, {}))
        # once get, it is cached
        vocab(self.portal)
        self.assertTrue(getattr(vocab, memPropName))
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2'])

        # change value but do not clean cache
        self.request.set('vocab_values', ('1', '2', '3'))
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2'])
        # clean vocabulary cache
        cleanVocabularyCacheFor("imio.helpers.testing.testingvocabulary")
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2', '3'])

        # every existing vocabularies can also be cleaned if nothing passed to cleanVocabularyCacheFor
        self.request.set('vocab_values', ('1', '2', '3', '4'))
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2', '3'])
        # clean vocabulary cache
        cleanVocabularyCacheFor("imio.helpers.testing.testingvocabulary")
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2', '3', '4'])

        # if cleanVocabularyCacheFor is called without parameter,
        # every registered vocabularies cache is cleaned
        self.request.set('vocab_values', ('1', '2', '3', '4', '5'))
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2', '3', '4'])
        cleanVocabularyCacheFor()
        self.assertEquals([term.token for term in vocab(self.portal)._terms],
                          ['1', '2', '3', '4', '5'])

    def test_cleanRamCache(self):
        """
        This helper method invalidates all ram.cache.
        """
        self.request.set('ramcached', 'a')
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # change value in REQUEST, as it is ram cached, it will still return 'a'
        self.request.set('ramcached', 'b')
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # ram.cache works as expected if param changes
        self.assertEquals(ramCachedMethod(self.portal, param='2'), 'b')
        # try again
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # now clean all caches, it will returns 'b'
        cleanRamCache()
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'b')

    def test_cleanRamCacheFor(self):
        """
        This helper method invalidates ram.cache for given method.
        """
        self.request.set('ramcached', 'a')
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # change value in REQUEST, as it is ram cached, it will still return 'a'
        self.request.set('ramcached', 'b')
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # ram.cache works as expected if param changes
        self.assertEquals(ramCachedMethod(self.portal, param='2'), 'b')
        # try again
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'a')
        # now clean cache, it will returns 'b'
        cleanRamCacheFor('imio.helpers.tests.test_cache.ramCachedMethod')
        self.assertEquals(ramCachedMethod(self.portal, param='1'), 'b')

    def test_cleanForeverCache(self):
        """Test the cache.cleanForeverCache function."""
        date = forever_cached_date()
        self.assertEqual(date, forever_cached_date())
        cleanForeverCache()
        new_date = forever_cached_date()
        self.assertNotEqual(date, new_date)
        self.assertEqual(new_date, forever_cached_date())

    def test_get_cachekey_volatile(self):
        """Helper method that adds a volatile on the portal storing current date."""
        method_name = 'My method'
        plone_utils = api.portal.get_tool('plone_utils')
        normalized_name = plone_utils.normalizeString(method_name)
        volatile_name = normalized_name
        self.assertIsNone(getattr(self.portal, volatile_name, None))
        # calling the method will set the volatile on the portal
        date = get_cachekey_volatile(method_name)
        self.assertTrue(isinstance(date, datetime))
        volatiles = getattr(self.portal, VOLATILE_ATTR, {})
        self.assertTrue(isinstance(volatiles.get(volatile_name), datetime))
        # calling it again will still return same date
        self.assertEquals(date, get_cachekey_volatile(method_name))
        # volatiles are not removed by tearDown, remove it now to avoid
        # test isolation issues with test test_invalidate_cachekey_volatile_for
        invalidate_cachekey_volatile_for(method_name)

    def test_get_cachekey_volatile_method(self):
        """Test behavior when method is passed to get_cachekey_volatile."""
        cache_chooser = getUtility(ICacheChooser)
        thecache = cache_chooser("")
        cache_data = thecache.ramcache._getStorage()._data
        # will initialize cachekey_volatile and pass method
        ramCachedMethod(self.portal, param='1', pass_volatile_method=True)
        self.assertTrue('imio.helpers.tests.test_cache.ramCachedMethod' in cache_data)
        volatiles = getattr(self.portal, VOLATILE_ATTR, {})
        self.assertTrue(volatiles["_methods_invalidation_mapping"]["test_cache.ramCachedMethod"])
        # when invalidating cache, the method is cleaned from ramcache
        invalidate_cachekey_volatile_for('test_cache.ramCachedMethod')
        self.assertFalse('imio.helpers.tests.test_cache.ramCachedMethod' in cache_data)
        self.assertEqual(volatiles["_methods_invalidation_mapping"], PersistentMapping())
        # if not passing method to get_cachekey_volatile then cache will not be cleanedup
        ramCachedMethod(self.portal, param='1', pass_volatile_method=False)
        self.assertTrue('imio.helpers.tests.test_cache.ramCachedMethod' in cache_data)
        invalidate_cachekey_volatile_for('test_cache.ramCachedMethod')
        # no cleanup, method still in cache storage
        self.assertTrue('imio.helpers.tests.test_cache.ramCachedMethod' in cache_data)

    def test_get_cachekey_volatile_ttl(self):
        """Check use of parameter "ttl"."""
        date = get_cachekey_volatile("method")
        time.sleep(2)
        self.assertEqual(date, get_cachekey_volatile("method"))
        self.assertNotEqual(date, get_cachekey_volatile("method", ttl=1))

    def test_invalidate_cachekey_volatile_for(self):
        """Helper method that will invalidate a given volatile."""
        method_name = 'My method'
        plone_utils = api.portal.get_tool('plone_utils')
        normalized_name = plone_utils.normalizeString(method_name)
        volatile_name = normalized_name
        self.assertIsNone(getattr(self.portal, volatile_name, None))
        # calling the method if volatile does not exist does not break
        invalidate_cachekey_volatile_for(method_name)
        # set it now
        first_date = get_cachekey_volatile(method_name)
        self.assertTrue(isinstance(first_date, datetime))
        invalidate_cachekey_volatile_for(method_name)
        self.assertIsNone(getattr(self.portal, volatile_name, None))
        # if get_cachekey_volatile is called and volatile does not exist,
        # it is created with datetime.now()
        second_date = get_cachekey_volatile(method_name)
        self.assertTrue(first_date < second_date)
        # when get_again=True, a new date is computed
        new_date = invalidate_cachekey_volatile_for(method_name, get_again=True)
        self.assertEqual(new_date, get_cachekey_volatile(method_name))

    def test_generate_params_key(self):
        """Test _generate_params_key function"""
        self.assertEqual(
            ((1, ), frozenset([('a', 2)])),
            _generate_params_key(1, a=2),
        )
        self.assertEqual(
            ((1, 2), frozenset([('a', 2), ('b', ())])),
            _generate_params_key(1, 2, a=2, b=[]),
        )

    def test_generate_key(self):
        """Test generate_key function"""
        self.assertEqual(
            'imio.helpers.cache.generate_key',
            generate_key(generate_key),
        )

        class Foo(object):

            def bar(self):
                pass

        self.assertEqual(
            'imio.helpers.tests.test_cache.Foo.bar',
            generate_key(Foo.bar),
        )

    def test_volatile_cache_without_parameters(self):
        """Helper cache @volatile_cache_without_parameters"""
        generate_key(volatile_without_parameters_cached)
        self.request.set('volatile_without_parameters_cached', 'a')
        self.assertEqual('a', volatile_without_parameters_cached(self.portal))
        self.request.set('volatile_without_parameters_cached', 'b')
        self.assertEqual('a', volatile_without_parameters_cached(self.portal))
        # Test invalidation
        invalidate_cachekey_volatile_for(
            generate_key(volatile_without_parameters_cached),
        )
        self.assertEqual('b', volatile_without_parameters_cached(self.portal))

    def test_volatile_cache_with_parameters(self):
        """Helper cache @volatile_cache_with_parameters"""
        self.request.set('volatile_with_parameters_cached', 'a')
        self.assertEqual(
            'a',
            volatile_with_parameters_cached(self.portal, 'a'),
        )
        self.request.set('volatile_with_parameters_cached', 'b')
        self.assertEqual(
            'a',
            volatile_with_parameters_cached(self.portal, 'a'),
        )
        self.assertEqual(
            'b',
            volatile_with_parameters_cached(self.portal, 'b'),
        )
        self.request.set('volatile_with_parameters_cached', 'c')
        self.assertEqual(
            'b',
            volatile_with_parameters_cached(self.portal, 'b'),
        )
        # Test invalidation
        invalidate_cachekey_volatile_for(
            generate_key(volatile_with_parameters_cached),
        )
        self.assertEqual(
            'c',
            volatile_with_parameters_cached(self.portal, 'a'),
        )
        self.assertEqual(
            'c',
            volatile_with_parameters_cached(self.portal, 'b'),
        )

    def test_obj_modified(self):
        """ """
        # init annotations
        folder = self.portal.folder
        ann = IAnnotations(folder)
        ann['test'] = 0
        transaction.commit()
        ann = folder.__annotations__
        modified = obj_modified(folder)
        # changed when modified
        folder.notifyModified()
        modified2 = obj_modified(folder)
        self.assertNotEqual(modified, modified2)
        # changed when annotation changed
        ann['test'] = 1
        transaction.commit()
        modified3 = obj_modified(folder)
        self.assertNotEqual(modified2, modified3)
        # also working when stored annotation is a dict and something changed into it
        ann['test2'] = PersistentMapping({'key': 'value1'})
        transaction.commit()
        modified4 = obj_modified(folder)
        self.assertNotEqual(modified3, modified4)
        ann['test2']['key'] = 'value2'
        transaction.commit()
        modified5 = obj_modified(folder)
        self.assertNotEqual(modified4, modified5)
        # by default result is a datetime, can be float
        self.assertEqual(type(obj_modified(folder)), datetime)
        self.assertEqual(type(obj_modified(folder, asdatetime=False)), float)


class TestCachedMethods(IntegrationTestCase):
    """
    Test methods that have been dynamically cached in __init__.
    """

    def setUp(self):
        super(TestCachedMethods, self).setUp()
        self.acl = self.portal['acl_users']
        api.user.create('a@b.be', 'user1', '12345', properties={'fullname': 'Stéphan Smith'})  # Returns MemberData
        self.user = self.acl.getUserById('user1')  # Returns PloneUser

    def test_getGroupsForPrincipal(self):
        self.skipTest('Temporary skipped')
        pgr = self.portal['portal_groups']
        self.assertListEqual(pgr.getGroupsForPrincipal(self.user),
                             ['AuthenticatedUsers'])
        pgr.addPrincipalToGroup('user1', 'Administrators')
        self.assertListEqual(pgr.getGroupsForPrincipal(self.user),
                             ['Administrators', 'AuthenticatedUsers'])
        api.group.add_user(groupname='Reviewers', user=self.user)
        self.assertListEqual(pgr.getGroupsForPrincipal(self.user),
                             ['Administrators', 'Reviewers', 'AuthenticatedUsers'])
        api.group.remove_user(groupname='Administrators', user=self.user)
        self.assertListEqual(pgr.getGroupsForPrincipal(self.user),
                             ['Reviewers', 'AuthenticatedUsers'])
        pgr.removePrincipalFromGroup('user1', 'Reviewers')
        self.assertListEqual(pgr.getGroupsForPrincipal(self.user),
                             ['AuthenticatedUsers'])

    def test_getRolesForPrincipal(self):
        self.assertIn(os.getenv('decorate_acl_methods', 'Nope'), ('True', 'true'))
        prm = self.acl['portal_role_manager']
        self.assertTupleEqual(prm.getRolesForPrincipal(self.user),
                              ('Member',))
        prm.assignRolesToPrincipal(('Member', 'Reviewer',), 'user1')
        self.assertTupleEqual(prm.getRolesForPrincipal(self.user),
                              ('Member', 'Reviewer'))
        prm.assignRoleToPrincipal('Contributor', 'user1')
        self.assertTupleEqual(prm.getRolesForPrincipal(self.user),
                              ('Member', 'Reviewer', 'Contributor'))
        prm.removeRoleFromPrincipal('Reviewer', 'user1')
        self.assertTupleEqual(prm.getRolesForPrincipal(self.user),
                              ('Member', 'Contributor'))

    def test__getGroupsForPrincipal(self):
        self.skipTest('Temporary skipped')
        pgr = self.portal['portal_groups']
        self.assertListEqual(self.acl._getGroupsForPrincipal(self.user),
                             ['AuthenticatedUsers'])
        pgr.addPrincipalToGroup('user1', 'Administrators')
        self.assertListEqual(self.acl._getGroupsForPrincipal(self.user),
                             ['Administrators', 'AuthenticatedUsers'])
        api.group.add_user(groupname='Reviewers', user=self.user)
        invalidate_cachekey_volatile_for('_users_groups_value', get_again=True)
        self.assertListEqual(self.acl._getGroupsForPrincipal(self.user),
                             ['Administrators', 'Reviewers', 'AuthenticatedUsers'])
        api.group.remove_user(groupname='Administrators', user=self.user)
        self.assertListEqual(self.acl._getGroupsForPrincipal(self.user),
                             ['Reviewers', 'AuthenticatedUsers'])
        pgr.removePrincipalFromGroup('user1', 'Reviewers')
        self.assertListEqual(self.acl._getGroupsForPrincipal(self.user),
                             ['AuthenticatedUsers'])

    def test_imio_global_cache(self):
        from zope.ramcache.ram import caches
        caches.clear()
        ramcache = queryUtility(IRAMCache)
        self.assertEqual(ramcache.getStatistics(), ())
        ramCachedMethod(self.portal, param='1', pass_volatile_method=True)
        stats = ramcache.getStatistics()
        self.assertEqual(len(stats), 1)
        self.assertEqual(stats[0]['path'], 'imio.helpers.tests.test_cache.ramCachedMethod')
        self.assertEqual(stats[0]['hits'], 0)
        self.assertTrue('older_date' in stats[0])
        # hit again
        ramCachedMethod(self.portal, param='1', pass_volatile_method=True)
        stats = ramcache.getStatistics()
        self.assertEqual(len(stats), 1)
        self.assertEqual(stats[0]['path'], 'imio.helpers.tests.test_cache.ramCachedMethod')
        self.assertEqual(stats[0]['hits'], 1)
        self.assertTrue('older_date' in stats[0])

    def test__users_groups_value_invalidation(self):
        """Invalidated whenever any operation on user:
           - user created;
           - user modified;
           - user deleted;
           - user added to group;
           - user removed from group;
           - group created;
           - group removed."""
        value1 = get_cachekey_volatile('_users_groups_value')
        # create user
        user = api.user.create(username='new_user', email='test@test.be')
        value2 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value1, value2)
        # modify user
        user.setMemberProperties({'email': 'test2@test.be'})
        value3 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value2, value3)
        # create group
        group = api.group.create('new_group')
        value4 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value3, value4)
        # modify group, WILL NOT BE INVALIDATED
        group.setGroupProperties({'title': 'Group title'})
        value5 = get_cachekey_volatile('_users_groups_value')
        self.assertEqual(value4, value5)
        # add user to group
        api.group.add_user(group=group, user=user)
        value6 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value5, value6)
        # remove user from group
        api.group.remove_user(group=group, user=user)
        value7 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value6, value7)
        # delete user
        api.user.delete(user=user)
        value8 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value7, value8)
        # delete group
        api.group.delete(group=group)
        value9 = get_cachekey_volatile('_users_groups_value')
        self.assertNotEqual(value8, value9)

    def test__listAllowedRolesAndUsers(self):
        pgr = self.portal['portal_groups']
        new_user = api.user.create(username='new_user', email='test@test.be')
        self.assertEqual(
            self.catalog._listAllowedRolesAndUsers(new_user),
            ['user:new_user', 'Member', 'Authenticated',
             'user:AuthenticatedUsers', 'Anonymous'])
        pgr.addPrincipalToGroup(new_user.getId(), 'Administrators')
        # get again, as we use "getGroups", it is stored on user instance
        new_user = api.user.get(new_user.getId())
        self.assertEqual(
            self.catalog._listAllowedRolesAndUsers(new_user),
            ['user:new_user', 'Member', 'Manager', 'Authenticated',
             'user:Administrators', 'user:AuthenticatedUsers', 'Anonymous'])
        # behaves correctly with a PloneUser because of id/getId
        plone_user1 = _getAuthenticatedUser(self.portal)
        self.assertTrue(isinstance(plone_user1, PloneUser))
        self.assertEqual(plone_user1.id, "acl_users")
        self.assertEqual(plone_user1.getId(), "test_user_1_")
        self.assertEqual(
            self.catalog._listAllowedRolesAndUsers(plone_user1),
            ['user:test_user_1_', 'Manager', 'Authenticated',
             'user:AuthenticatedUsers', 'Anonymous'])
        login(self.portal, new_user.getId())
        plone_user2 = _getAuthenticatedUser(self.portal)
        self.assertTrue(isinstance(plone_user2, PloneUser))
        self.assertEqual(plone_user2.id, "acl_users")
        self.assertEqual(plone_user2.getId(), "new_user")
        self.assertEqual(
            self.catalog._listAllowedRolesAndUsers(plone_user2),
            ['user:new_user', 'Member', 'Manager', 'Authenticated',
             'user:Administrators', 'user:AuthenticatedUsers', 'Anonymous'])
        # as Anonymous
        logout()
        anon_user = _getAuthenticatedUser(self.portal)
        self.assertEqual(
            self.catalog._listAllowedRolesAndUsers(anon_user),
            ['Anonymous'])



# -*- coding: utf-8 -*-

from plone import api
from plone.dexterity.fti import DexterityFTI
from Products.GenericSetup.utils import importObjects

import logging


logger = logging.getLogger('imio.helpers.setup')


def load_workflow_from_package(wkf_name, profile_id):
    """Loads a workflow from his xml definition.
    :param wkf_name: workflow id
    :param profile_id: package profile id
    :return: status as boolean
    """
    wkf_tool = api.portal.get_tool('portal_workflow')
    wkf_obj = wkf_tool.get(wkf_name)
    if wkf_obj is None:
        logger.error("Cannot find '{}' workflow name in portal".format(wkf_name))
        return False
    ps_tool = api.portal.get_tool('portal_setup')
    try:
        context = ps_tool._getImportContext(profile_id, True)
    except KeyError:
        logger.error("Cannot find '{}' profile id".format(profile_id))
        return False
    # ps_tool.applyContext(context)  # necessary ?
    importObjects(wkf_obj, 'workflows/', context)
    if wkf_obj._p_changed is False:
        logger.error("Could not update '{}' using profile '{}'".format(wkf_name, profile_id))
        return False
    return True


def load_type_from_package(type_name, profile_id, purge_actions=False):
    """Loads a portal_type from his xml definition.
    :param type_name: portal_type id
    :param profile_id: package profile id
    :return: status as boolean
    """
    types_tool = api.portal.get_tool('portal_types')
    portal_type = types_tool.get(type_name)
    if portal_type is None:
        logger.error("Cannot find '{}' portal_type name in portal".format(type_name))
        return False
    ps_tool = api.portal.get_tool('portal_setup')
    try:
        context = ps_tool._getImportContext(profile_id, True)
    except KeyError:
        logger.error("Cannot find '{}' profile id".format(profile_id))
        return False

    # special case for DX FTI, _should_purge is set to False or it fails when purging
    if isinstance(portal_type, DexterityFTI):
        context._should_purge = False

    if purge_actions:
        # remove actions so it is reloaded in correct order
        portal_type._actions = ()

    # ps_tool.applyContext(context)  # necessary ?
    importObjects(portal_type, 'types/', context)
    if portal_type._p_changed is False:
        logger.error("Could not update '{}' using profile '{}'".format(type_name, profile_id))
        return False
    return True



# -*- coding: utf-8 -*-
from copy import deepcopy
from datetime import datetime
from future.builtins import zip
from imio.helpers.content import safe_encode
from imio.pyutils.utils import letters_sequence
from Products.CMFPlone.utils import safe_unicode

import os
import re


def clean_value(value, isep=u'\n', strip=u' ', patterns=(), osep=None):
    """Clean unicode multiline value

    :param value: input string
    :param isep: input separator
    :param strip: chars to strip on each "line"
    :param patterns: tuple line patterns list to remove ("line" evaluation) (tuple = search, replace)
    :param osep: output separator
    :return: string
    """
    parts = []
    if not value:
        return value
    if osep is None:
        osep = isep
    for part in value.split(isep):
        part = part.strip(strip)
        for pattern, replace in patterns:
            part = re.sub(pattern, replace, part, flags=re.U)
        if part:
            parts.append(part)
    return osep.join(parts)


def filter_keys(item, keys, unfound=None):
    """Return a copy of item with only given keys.

    :param item: yielded item (dict)
    :param keys: keys to keep
    :param unfound: unfound value (default None)
    :return: filtered item
    """
    if not keys:
        return deepcopy(item)
    new_item = {}
    for key in keys:
        new_item[key] = item.get(key, unfound)
    return new_item


def get_correct_id(obj, oid, with_letter=False):
    """ Modify an id already existing in obj.

    :param obj: plone obj or dict or list
    :param oid: id to check
    :param with_letter: add a letter prefix
    :return: unique id
    """
    original = oid
    i = 1
    while oid in obj:
        sfx = with_letter and letters_sequence(i) or i
        oid = u'{}-{}'.format(original, sfx)
        i += 1
    return oid


def get_correct_path(portal, path):
    """ Check if a path already exists on obj """
    original = path
    i = 1
    while portal.unrestrictedTraverse(path, default=None) is not None:  # path exists
        path = '{}-{:d}'.format(original, i)
        i += 1
    return path


def get_obj_from_path(root, item={}, path_key='_path', path=None):
    """Gets object from path stored in item (dic) or from given path

    :param root: root object from where the search is done
    :param item: dict containing the path
    :param path_key: dict key to access the path value
    :param path: path value
    :return: the object or None if not found
    """
    if not path:
        path = item.get(path_key)
    if not path:
        return None
    path = safe_encode(path)
    if path.startswith('/'):
        path = path[1:]
    try:
        return root.unrestrictedTraverse(path)
    except (KeyError, AttributeError):
        return None


def get_main_path(path='', subpath=''):
    """Return path/subpath if it exists. If path is empty, return buildout path."""
    if not path:
        # Are we in a classic buildout
        instance_home = os.getenv('INSTANCE_HOME')
        match = re.match('(.+)/parts/.+', instance_home)
        if match:
            path = match.group(1)
        else:
            path = os.getenv('PWD')
    if subpath:
        path = os.path.join(path, subpath)
    if os.path.exists(path):
        return path
    raise Exception("Path '{}' doesn't exist".format(path))


def pool_tuples(iterable, pool_len=2, e_msg=''):
    """Returns the iterable as tuples

    :param iterable: a list or tuple
    :param pool_len: len of output pool tuples
    :param e_msg: prefix message when the iterable has not a correct len
    :return: an iterator giving pool tuples
    """
    if not iterable:
        return iterable
    if len(iterable) % pool_len:
        raise ValueError("{}: the given iterable must contain a multiple of {} elements: value = '{}'".format(
                         e_msg, pool_len, iterable))
    l_iter = iter(iterable)
    args = [l_iter for x in range(0, pool_len)]
    return zip(*args)


def relative_path(portal, fullpath, with_slash=True):
    """Returns relative path following given portal.

    :param portal: leading object to remove from path
    :param fullpath: path to update
    :param with_slash: keep leadind slash
    :return: new path relative to portal object parameter
    """
    portal_path = '/'.join(portal.getPhysicalPath())  # not unicode, brain.getPath is also encoded
    if not fullpath.startswith(portal_path):
        return fullpath
    shift = not with_slash and 1 or 0
    return fullpath[len(portal_path) + shift:]


def key_val(key, dic):
    """ Return a dic value or the key

    :param key: given key
    :param dic: dict
    :return: value or key
    """
    return dic.get(key, key)


def str_to_bool(item, key, log_method, **log_params):
    """Changed to bool the text value of item[key].

    :param item: yielded item (usually dict)
    :param key: dict key
    :param log_method: special log method (as log_error in imio.transmogrifier.iadocs.utils)
    :param log_params: log method keyword parameters
    :return: the boolean value
    """
    if log_params is None:
        log_params = {}
    if item[key] in (u'True', u'true'):
        return True
    elif item[key] in (u'False', u'false'):
        return False
    try:
        return bool(int(item[key] or 0))
    except Exception:  # noqa
        log_method(item, u"Cannot change '{}' key value '{}' to bool".format(key, item[key]), **log_params)
    return False


def split_text(text, length):
    """Split text on a word boundary at max length"""
    if text is None:
        return u'', u''
    part1 = text = safe_unicode(text)
    part2 = u''
    if len(text) > length:
        part1 = text[:length]
        part2 = text[length:]
        if part2[:1] != u' ':  # cut in a word
            s_i = part1.rfind(u' ')
            if s_i > length / 2:  # space after the middle of the part1
                part1 = part1[:s_i + 1]
                part2 = text[s_i + 1:]
    return part1, part2


def str_to_date(item, key, log_method, fmt='%Y/%m/%d', can_be_empty=True, as_date=True, min_val=None, max_val=None,
                **log_params):
    """Changed to date or datetime the text value of item[key]

    :param item: yielded item (usually dict)
    :param key: dict key
    :param log_method: special log method (as log_error in imio.transmogrifier.iadocs.utils)
    :param fmt: formatting date string
    :param can_be_empty: value can be empty or None
    :param as_date: return a date, otherwise a datetime
    :param min_val: minimal date
    :param max_val: maximum date
    :param log_params: log method keyword parameters
    :return: the date or datetime value
    """
    val = item.get(key)
    if not val and can_be_empty:
        return None
    try:
        dt = datetime.strptime(val, fmt)
        if as_date:
            dt = dt.date()
        if min_val and dt < min_val:
            raise(ValueError(u"Given date '{}' < minimal value '{}' => set to None".format(val, min_val)))
        if max_val and dt > max_val:
            raise (ValueError(u"Given date '{}' > maximum value '{}' => set to None".format(val, max_val)))
    except ValueError as ex:
        log_method(item, u"not a valid date '{}' in key '{}': {}".format(val, key, ex), **log_params)
        return None
    return dt



# -*- coding: utf-8 -*-

from imio.helpers.testing import IntegrationTestCase
from ZPublisher.tests.testHTTPRequest import HTTPRequestTests

import imio.helpers.converters  # noqa


class TestConverters(IntegrationTestCase, HTTPRequestTests):

    def test_imiohelpers_json_converter(self):
        inputs = (
            ('data:json', '{"key1": "value1", "key2": "value2"}'),
            ('data2:json', '{"key3": "value3", "key4": "value4"}'), )
        req = self._processInputs(inputs)
        self.assertEqual(req.form,
                         {'data': {u'key2': u'value2', u'key1': u'value1'},
                          'data2': {u'key3': u'value3', u'key4': u'value4'}})

    def test_imiohelpers_json_converter_wrong_format(self):
        # missing " after value1
        inputs = (
            ('data:json', '{"key1": "value1, "key2": "value2"}'), )
        with self.assertRaises(ValueError) as cm:
            self._processInputs(inputs)
        self.assertEqual(cm.exception.message,
                         'Invalid json \'{"key1": "value1, "key2": "value2"}\'')


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    # change prefix to avoid executing every tests of HTTPRequestTests
    suite.addTest(makeSuite(TestConverters, prefix='test_imiohelpers_'))
    return suite



# -*- coding: utf-8 -*-

from App.config import _config
from imio.helpers.security import check_zope_admin
from imio.helpers.security import fplog
from imio.helpers.security import generate_password
from imio.helpers.security import get_environment
from imio.helpers.security import get_user_from_criteria
from imio.helpers.security import get_zope_root
from imio.helpers.security import is_develop_environment
from imio.helpers.security import set_site_from_package_config
from imio.helpers.security import setup_logger
from imio.helpers.testing import IntegrationTestCase
from OFS.Application import Application
from plone import api
from plone.app.testing import login
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import _checkPermission
from Products.CMFPlone.Portal import PloneSite

import os
import re


class TestSecurityModule(IntegrationTestCase):
    """
    Test all helper methods of security module.
    """

    def setUp(self):
        if 'IS_DEV_ENV' in os.environ.keys():
            del os.environ['IS_DEV_ENV']
        if 'ENV' in os.environ.keys():
            del os.environ['ENV']
        self.portal = self.layer['portal']

    def test_fplog(self):
        fplog('test_action', 'test_extras')

    def test_is_develop_environment(self):
        os.environ['IS_DEV_ENV'] = 'false'
        self.assertFalse(is_develop_environment())
        os.environ['IS_DEV_ENV'] = 'true'
        self.assertTrue(is_develop_environment())

    def test_get_environment(self):
        os.environ['ENV'] = 'staging'
        self.assertFalse(is_develop_environment())
        self.assertEqual(get_environment(), 'staging')
        os.environ['ENV'] = 'dev'
        self.assertEqual(get_environment(), 'dev')
        self.assertTrue(is_develop_environment())

    def test_generate_password(self):
        pwd = generate_password(length=10, digits=1, upper=1, lower=1, special=1, readable=False)
        self.assertEqual(len(pwd), 10)
        self.assertIsNotNone(re.search(r'\d', pwd))
        self.assertIsNotNone(re.search(r'[A-Z]', pwd))
        self.assertIsNotNone(re.search(r'[a-z]', pwd))
        self.assertIsNotNone(re.search(r'[!#$%&*+<=>?@-]', pwd))

    def test_get_user_from_criteria(self):
        api.user.create('elementary@mail.be', 'ssmith', '12345', properties={'fullname': 'Stéphan Smith'})
        api.user.create('teacher@school.be', 'teacher', '12345', properties={'fullname': 'Paul Smith'})
        self.assertEqual(len(get_user_from_criteria(self.portal, email='unknown')), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname='unknown')), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid='unknown')), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, email='mentary')), 1)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname='Stéph')), 1)
        self.assertEqual(len(get_user_from_criteria(self.portal, email='.be')), 2)
        self.assertEqual(len(get_user_from_criteria(self.portal, email='')), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname='Smith')), 2)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname='')), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid='')), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid='teach')), 1)

    def test_setup_logger(self):
        # just call it to check that it is not broken
        self.assertIsNone(setup_logger())

    def test_get_zope_root(self):
        app = get_zope_root()
        self.assertIsInstance(app, Application)

    def test_set_site_from_package_config(self):
        # no config
        self.assertIsNone(set_site_from_package_config('imio.helpers'))
        # we add config
        _config.product_config = {'imio.helpers': {'plone-path': 'plone'}}
        site = set_site_from_package_config('imio.helpers')
        self.assertIsInstance(site, PloneSite)

    def test_check_zope_admin(self):
        # create and login as siteadmin
        self.portal.portal_membership.addMember('siteadmin', 'siteadmin', ['Manager'], [])
        login(self.portal, 'siteadmin')
        self.assertTrue(_checkPermission(ManagePortal, self.portal))
        self.assertFalse(check_zope_admin())
        # login as Zope admin
        login(self.portal.aq_parent, 'admin')
        self.assertTrue(_checkPermission(ManagePortal, self.portal))
        self.assertTrue(check_zope_admin())



# -*- coding: utf-8 -*-
from imio.helpers.testing import IntegrationTestCase
from imio.helpers.transmogrifier import clean_value
from imio.helpers.transmogrifier import filter_keys
from imio.helpers.transmogrifier import get_correct_id
from imio.helpers.transmogrifier import get_correct_path
from imio.helpers.transmogrifier import get_main_path
from imio.helpers.transmogrifier import get_obj_from_path
from imio.helpers.transmogrifier import key_val
from imio.helpers.transmogrifier import pool_tuples
from imio.helpers.transmogrifier import relative_path
from imio.helpers.transmogrifier import split_text
from imio.helpers.transmogrifier import str_to_bool
from imio.helpers.transmogrifier import str_to_date

import datetime
import os


# logger simulation for some tests
def logger(item, msg):
    item['errors'] += 1


class TestTesting(IntegrationTestCase):

    def test_clean_value(self):
        self.assertEqual(clean_value(None), None)
        self.assertEqual(clean_value(u'  strip  '), u'strip')
        self.assertEqual(clean_value(u' | strip  ', strip=u' |'), u'strip')
        self.assertEqual(clean_value(u' strip  \n  strip  '), u'strip\nstrip')
        self.assertEqual(clean_value(u' strip  \n  strip  ', osep=u'\r\n'), u'strip\r\nstrip')
        self.assertEqual(clean_value(u' strip  |  strip  ', isep=u'|'), u'strip|strip')
        self.assertEqual(clean_value(u'  \n strip  \n '), u'strip')
        self.assertEqual(clean_value(u' strip  \n "', strip=u' "'), u'strip')
        self.assertEqual(clean_value(u' strip  \n "', patterns=[(r'"', u'')]), u'strip')
        self.assertEqual(clean_value(u' strip  \n "\'"', patterns=[(r'^["\']+$', u'')]), u'strip')
        self.assertEqual(clean_value(u' strip  \n ', patterns=[]), u'strip')
        self.assertEqual(clean_value(u' strip  \n line2', patterns=[], osep=u'-'), u'strip-line2')
        self.assertEqual(clean_value(u' stri    p  \n line  2', patterns=[(r'\s{2,}', u' ')], osep=u'-'),
                         u'stri p-line 2')

    def test_get_correct_id(self):
        self.assertEquals(get_correct_id(self.portal, 'abcde'), 'abcde')
        self.assertIn('folder', self.portal.objectIds())
        self.assertEquals(get_correct_id(self.portal, 'folder'), 'folder-1')
        self.assertEquals(get_correct_id(self.portal, 'folder', with_letter=True), 'folder-a')
        self.assertEquals(get_correct_id(self.portal.folder, 'abcde'), 'abcde')
        self.portal.folder.invokeFactory('Document', id='abcde', title='Document')
        self.assertEquals(get_correct_id(self.portal.folder, 'abcde'), 'abcde-1')
        self.assertEquals(get_correct_id(self.portal.folder, 'abcde', True), 'abcde-a')
        self.portal.folder.invokeFactory('Document', id='abcde-1', title='Document')
        self.assertEquals(get_correct_id(self.portal.folder, 'abcde'), 'abcde-2')

    def test_get_correct_id_dic(self):
        lst = ['a']
        self.assertEquals(get_correct_id(lst, 'z'), 'z')
        self.assertEquals(get_correct_id(lst, 'a'), 'a-1')
        self.assertEquals(get_correct_id(lst, 'a', True), 'a-a')
        lst.extend(['a-1', 'a-a'])
        self.assertEquals(get_correct_id(lst, 'a'), 'a-2')
        self.assertEquals(get_correct_id(lst, 'a', True), 'a-b')
        lst.extend(['a-{}'.format(lt) for lt in 'bcdefghijklmnopqrstuvwxyz'])
        self.assertEquals(get_correct_id(lst, 'a', True), 'a-aa')

    def test_get_correct_path(self):
        self.assertEquals(get_correct_path(self.portal, 'abcde'), 'abcde')
        self.assertIn('folder', self.portal.objectIds())
        self.assertEquals(get_correct_path(self.portal, 'folder'), 'folder-1')
        self.assertEquals(get_correct_path(self.portal, 'folder/abcde'), 'folder/abcde')
        self.portal.folder.invokeFactory('Document', id='abcde', title='Document')
        self.assertEquals(get_correct_path(self.portal, 'folder/abcde'), 'folder/abcde-1')
        self.portal.folder.invokeFactory('Document', id='abcde-1', title='Document')
        self.assertEquals(get_correct_path(self.portal, 'folder/abcde'), 'folder/abcde-2')

    def test_filter_keys(self):
        dic = {'a': 1, 'b': 2, 'c': 3}
        self.assertListEqual(['a', 'b', 'c'], sorted(filter_keys(dic, []).keys()))
        self.assertListEqual([1, 2, 3], sorted(filter_keys(dic, []).values()))
        self.assertListEqual(['a', 'b', 'c'], sorted(filter_keys(dic, ['a', 'b', 'c']).keys()))
        self.assertListEqual([1, 2, 3], sorted(filter_keys(dic, ['a', 'b', 'c']).values()))
        self.assertListEqual(['a', 'c'], sorted(filter_keys(dic, ['a', 'c']).keys()))
        self.assertListEqual([1, 3], sorted(filter_keys(dic, ['a', 'c']).values()))
        del dic['c']
        self.assertListEqual(['a', 'b', 'c'], sorted(filter_keys(dic, ['a', 'b', 'c']).keys()))
        self.assertListEqual([None, 1, 2], sorted(filter_keys(dic, ['a', 'b', 'c']).values()))
        self.assertListEqual([1, 2, ''], sorted(filter_keys(dic, ['a', 'b', 'c'], unfound='').values()))

    def test_get_obj_from_path(self):
        folder = self.portal.folder
        self.assertEqual(get_obj_from_path(self.portal, path='/folder'), folder)
        self.assertEqual(get_obj_from_path(self.portal, path='folder'), folder)
        self.assertEqual(get_obj_from_path(self.portal, path=u'folder'), folder)
        self.assertEqual(get_obj_from_path(self.portal, {'key': 'folder'}, path_key='key'), folder)
        self.assertIsNone(get_obj_from_path(self.portal, {}, path_key='unknown_key'))
        self.assertIsNone(get_obj_from_path(self.portal, path='in_va_lid'))

    def test_get_main_path(self):
        orig_home = os.getenv('INSTANCE_HOME')
        orig_pwd = os.getenv('PWD')
        self.assertEquals(get_main_path('/etc'), '/etc')
        os.environ['INSTANCE_HOME'] = '/etc/parts/xx'
        self.assertEquals(get_main_path(), '/etc')
        os.environ['INSTANCE_HOME'] = '/etc/abcd'
        os.environ['PWD'] = '/myhome'
        self.assertRaisesRegexp(Exception, u"Path '/myhome' doesn't exist", get_main_path)  # path doesn't exist
        self.assertRaisesRegexp(Exception, u"Path '/myhome/toto' doesn't exist", get_main_path, subpath='toto')
        self.assertEquals(get_main_path('/', 'etc'), '/etc')
        os.environ['INSTANCE_HOME'] = orig_home
        os.environ['PWD'] = orig_pwd

    def test_key_val(self):
        dic = {'key': 'val'}
        self.assertEqual(key_val('key', dic), 'val')
        self.assertEqual(key_val('unknown', dic), 'unknown')

    def test_pool_tuples(self):
        lst = [1, 2, 3, 4, 5, 6]
        self.assertEqual(pool_tuples(None, 1), None)
        self.assertListEqual(pool_tuples([], 1), [])
        self.assertListEqual([t for t in pool_tuples(lst, 1)], [(1,), (2,), (3,), (4,), (5,), (6,)])
        self.assertListEqual([t for t in pool_tuples(lst, 2)], [(1, 2), (3, 4), (5, 6)])
        self.assertListEqual([t for t in pool_tuples(lst, 3)], [(1, 2, 3), (4, 5, 6)])
        self.assertListEqual([t for t in pool_tuples(lst, 6)], [(1, 2, 3, 4, 5, 6)])
        with self.assertRaises(Exception) as cm:
            pool_tuples([1, 2, 3], 2, e_msg=u'My pool')
        self.assertEqual(str(cm.exception), "My pool: the given iterable must contain a multiple of 2 elements: "
                         "value = '[1, 2, 3]'")

    def test_relative_path(self):
        self.assertEquals(relative_path(self.portal, '/plone/directory'), '/directory')
        self.assertEquals(relative_path(self.portal, '/plone/directory', False), 'directory')
        self.assertEquals(relative_path(self.portal, '/alone/directory'), '/alone/directory')

    def test_split_text(self):
        self.assertTupleEqual(split_text(None, 50), (u'', u''))
        self.assertTupleEqual(split_text('aa bb cc dd éé', 50), (u'aa bb cc dd éé', u''))
        self.assertTupleEqual(split_text(u'aa bb cc dd ee', 50), (u'aa bb cc dd ee', u''))
        self.assertTupleEqual(split_text(u'aa bb cc dd ee', 0), (u'', u'aa bb cc dd ee'))
        self.assertTupleEqual(split_text(u'aa bb cc dd ee', 7), (u'aa bb ', u'cc dd ee'))
        self.assertTupleEqual(split_text(u'aa bb cc dd ee', 8), (u'aa bb cc', u' dd ee'))
        self.assertTupleEqual(split_text(u'aa bb cc dd ee', 9), (u'aa bb cc ', u'dd ee'))
        # space before the middle
        self.assertTupleEqual(split_text(u'aa bbccddee', 9), (u'aa bbccdd', u'ee'))

    def test_str_to_bool(self):
        dic = {'True': 'True', 'true': 'true', 'False': 'False', 'false': 'false',
               '0': '0', 0: 0, '1': '1', 123: 123,
               'a': 'a', 'errors': 0}
        self.assertTrue(str_to_bool(dic, 'True', logger))
        self.assertTrue(str_to_bool(dic, 'true', logger))
        self.assertFalse(str_to_bool(dic, 'False', logger))
        self.assertFalse(str_to_bool(dic, 'false', logger))
        self.assertFalse(str_to_bool(dic, '0', logger))  # using fake log method to test error
        self.assertFalse(str_to_bool(dic, 0, logger))
        self.assertTrue(str_to_bool(dic, '1', logger))
        self.assertTrue(str_to_bool(dic, 123, logger))
        self.assertEquals(dic['errors'], 0)
        self.assertFalse(str_to_bool(dic, 'a', logger))
        self.assertEquals(dic['errors'], 1)

    def test_str_to_date(self):
        dic = {1: '2023/03/03', 2: '2023/03/03 09:29', 'errors': 0}
        # str_to_date(item, key, log_method, fmt='%Y/%m/%d', can_be_empty=True, as_date=True, **log_params):
        self.assertIsNone(str_to_date(dic, 'bad_key', logger))
        self.assertRaises(TypeError, str_to_date, dic, 'bad_key', logger, can_be_empty=False)
        self.assertIsNone(str_to_date(dic, 1, logger, min_val=datetime.date(2023, 4, 1)))
        self.assertEquals(dic['errors'], 1)
        self.assertIsNone(str_to_date(dic, 1, logger, max_val=datetime.date(2022, 4, 1)))
        self.assertEquals(dic['errors'], 2)
        self.assertIsNone(str_to_date(dic, 1, logger, min_val=datetime.date(2023, 4, 1),
                                      max_val=datetime.date(2022, 4, 1)))
        self.assertEquals(dic['errors'], 3)
        ret = str_to_date(dic, 1, logger)
        self.assertIsInstance(ret, datetime.date)
        self.assertEqual(str(ret), '2023-03-03')
        ret = str_to_date(dic, 2, logger, fmt='%Y/%m/%d %H:%M', as_date=False)
        self.assertIsInstance(ret, datetime.datetime)
        self.assertEqual(str(ret), '2023-03-03 09:29:00')
        ret = str_to_date(dic, 2, logger, fmt='%Y/%m/%d %M:%H', as_date=False)
        self.assertEquals(dic['errors'], 4)



# -*- coding: utf-8 -*-
from Acquisition import aq_base
from html import escape
from os import path
from plone import api
from plone.app.imaging.scale import ImageScale
from plone.outputfilters.browser.resolveuid import uuidToURL
from plone.outputfilters.filters.resolveuid_and_caption import ResolveUIDAndCaptionFilter
from Products.CMFPlone.utils import safe_unicode
from six.moves.html_parser import HTMLParser
from zExceptions import NotFound
from zope.container.interfaces import INameChooser

import base64
import cgi
import logging
import lxml.html
import os
import pkg_resources
import types
import urllib


logger = logging.getLogger('imio.helpers:xhtml')

try:
    HAS_CKEDITOR = True
    pkg_resources.get_distribution('collective.ckeditor')
except pkg_resources.DistributionNotFound:
    HAS_CKEDITOR = False


CLASS_TO_LAST_CHILDREN_NUMBER_OF_CHARS_DEFAULT = 240
# Xhtml tags that may contain content
CONTENT_TAGS = ('p', 'div', 'strong', 'span', 'strike', 'i', 'em', 'u',
                'small', 'mark', 'del', 'ins', 'sub', 'sup', 'ul', 'li')


def xhtmlContentIsEmpty(xhtmlContent, tagWithAttributeIsNotEmpty=True):
    '''This method checks if given p_xhtmlContent will produce someting on rendering.
       p_xhtmlContent can either be a string or already a lxml.html element.
       If p_tagWithAttributeIsNotEmpty is True, a tag without text but with an attribute will
       be considered not empty.'''

    # first check if xhtmlContent is not simply None or so
    isStr = isinstance(xhtmlContent, types.StringType) or isinstance(xhtmlContent, types.NoneType)
    if isStr and (not xhtmlContent or not xhtmlContent.strip()):
        return True

    if isStr:
        xhtmlContent = "<special_tag>%s</special_tag>" % xhtmlContent
        tree = lxml.html.fromstring(safe_unicode(xhtmlContent))
    else:
        tree = xhtmlContent

    def _isEmpty(child):
        # consider that child as children except for some tag type
        childAsChildren = [subchild for subchild in child.getchildren() if subchild.tag not in ('br', )]
        return not bool(child.text_content().strip()) and \
            not bool(tagWithAttributeIsNotEmpty and child.attrib) and \
            not bool(childAsChildren)

    if tree.tag == 'special_tag':
        if tree.getchildren():
            for child in tree.getchildren():
                if not _isEmpty(child):
                    return False
            return True

    # if xhtmlContent renders text or has attributes or has children, we consider it not empty
    return _isEmpty(tree)


def removeBlanks(xhtmlContent, pretty_print=False):
    '''This method will remove any blank line in p_xhtmlContent.'''
    tree = _turnToLxmlTree(xhtmlContent)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtmlContent

    for el in tree.getchildren():
        # el can be a subtree, like <ul><li>...</li></ul> we must consider entire rendering of it
        if xhtmlContentIsEmpty(el, tagWithAttributeIsNotEmpty=False):
            el.getparent().remove(el)
    # only return children of the <special_tag>
    return ''.join([lxml.html.tostring(x,
                                       encoding='ascii',
                                       pretty_print=pretty_print,
                                       method='html')
                    for x in tree.iterchildren()])


def replace_content(xhtml_content,
                    css_class,
                    new_content=u"",
                    new_content_link={},
                    pretty_print=False):
    '''This method will get tags using given p_css_class and
       replace it's content with given p_content.'''
    tree = _turnToLxmlTree(xhtml_content)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtml_content

    new_content = safe_unicode(new_content)

    from lxml.cssselect import CSSSelector
    selector = CSSSelector('.' + css_class)
    elements = selector(tree)
    for main_elt in elements:
        # will find every contained elements
        # store itered elements as we add new elements,
        # iteration ignores last elements
        elts = [elt for elt in main_elt.iter()]
        for elt in elts:
            is_main_elt = elt == main_elt
            # make sure elt does not contain sub element
            for sub_elt in elt.getchildren():
                elt.remove(sub_elt)
            if is_main_elt:
                if new_content_link:
                    elt.text = u""
                    new_elt = lxml.html.Element("a")
                    new_elt.attrib["href"] = new_content_link["href"]
                    new_elt.attrib["title"] = new_content_link.get("title", u"")
                    new_elt.text = new_content or u""
                    elt.append(new_elt)
                else:
                    elt.text = new_content or u""
            else:
                elt.tag = "span"
                elt.text = u""
                elt.tail = u""

    # only return children of the <special_tag>
    return ''.join([lxml.html.tostring(x,
                                       encoding='ascii',
                                       pretty_print=pretty_print,
                                       method='html')
                    for x in tree.iterchildren()])


def _turnToLxmlTree(xhtmlContent):
    if not xhtmlContent or not xhtmlContent.strip():
        return xhtmlContent

    # surround xhtmlContent with a special tag so we are sure that tree is always
    # a list of children of this special tag
    tree = lxml.html.fromstring(safe_unicode("<special_tag>%s</special_tag>" % xhtmlContent))
    children = tree.getchildren()
    if not children:
        return xhtmlContent
    return tree


def addClassToContent(xhtmlContent, css_class, pretty_print=False):
    """Add css class attribute p_css_class to every CONTENT_TAGS of p_xhtmlContent."""
    if isinstance(xhtmlContent, lxml.html.HtmlElement):
        children = [xhtmlContent]
    else:
        tree = _turnToLxmlTree(xhtmlContent)
        if not isinstance(tree, lxml.html.HtmlElement):
            return xhtmlContent
        children = tree.getchildren()

    for child in children:
        if child.tag in CONTENT_TAGS:
            exisingCssClass = child.attrib.get('class', '')
            if exisingCssClass:
                cssClass = '{0} {1}'.format(css_class, exisingCssClass)
            else:
                cssClass = css_class
            child.attrib['class'] = cssClass

    # use encoding to 'ascii' so HTML entities are translated to something readable
    res = ''.join([lxml.html.tostring(x,
                                      encoding='ascii',
                                      pretty_print=pretty_print,
                                      method='html') for x in children])
    return res


def addClassToLastChildren(xhtmlContent,
                           classNames={'p': 'ParaKWN',
                                       'div': 'ParaKWN',
                                       'strong': '',
                                       'span': '',
                                       'strike': '',
                                       'i': '',
                                       'em': '',
                                       'u': '',
                                       'small': '',
                                       'mark': '',
                                       'del': '',
                                       'ins': '',
                                       'sub': '',
                                       'sup': '',
                                       'ol': 'podNumberItemKeepWithNext',
                                       'ul': 'podBulletItemKeepWithNext',
                                       'li': 'podItemKeepWithNext'},
                           numberOfChars=CLASS_TO_LAST_CHILDREN_NUMBER_OF_CHARS_DEFAULT,
                           pretty_print=False):
    '''This method will add a class attribute adding class correspondig to tag given in p_classNames
       to the last tags of given p_xhtmlContent.
       It only consider given p_classNames keys which are text formatting tags and will define the class
       on last tags until it contains given p_numberOfChars number of characters.
    '''
    tree = _turnToLxmlTree(xhtmlContent)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtmlContent

    children = tree.getchildren()

    def adaptTree(children, managedNumberOfChars=0):
        """
          Recursive method that walk the children and subchildren and adapt what necessary.
        """
        # apply style on last element until we reached necessary numberOfChars or we encounter
        # a tag not in CONTENT_TAGS or we do not have a tag...
        i = 1
        stillNeedToAdaptPreviousChild = True
        numberOfChildren = len(children)
        while stillNeedToAdaptPreviousChild and i <= numberOfChildren:
            child = children[-i]
            if child.tag not in CONTENT_TAGS and not child.getchildren():
                stillNeedToAdaptPreviousChild = False
            else:
                cssClass = classNames.get(child.tag, '')
                if cssClass:
                    addClassToContent(child, cssClass)
                managedNumberOfChars += child.text_content() and len(child.text_content()) or 0
                subchildren = child.getchildren()
                if subchildren:
                    # recursion is done here
                    managedNumberOfChars = adaptTree(subchildren, managedNumberOfChars=managedNumberOfChars)

                if managedNumberOfChars >= numberOfChars:
                    stillNeedToAdaptPreviousChild = False
                i = i + 1
        return managedNumberOfChars

    # call recursive method 'adaptTree' that whill adapt children and subchildren
    adaptTree(children)

    # use encoding to 'ascii' so HTML entities are translated to something readable
    res = ''.join([lxml.html.tostring(x,
                                      encoding='ascii',
                                      pretty_print=pretty_print,
                                      method='html') for x in tree.iterchildren()])
    return res


def markEmptyTags(xhtmlContent,
                  markingClass='highlightBlankRow',
                  tagTitle='',
                  onlyAtTheEnd=False,
                  tags=('p', 'div', ),
                  pretty_print=False):
    '''This will add a CSS class p_markingClass to tags of the given p_xhtmlContent
       that are empty.  If p_onlyAtTheEnd is True, it will only mark empty rows that are
       ending the XHTML content.'''
    tree = _turnToLxmlTree(xhtmlContent)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtmlContent

    childrenToMark = []
    # find children to mark, aka empty tags taking into account p_onlyAtTheEnd
    for child in tree.getchildren():
        if child.tag in tags and xhtmlContentIsEmpty(child, tagWithAttributeIsNotEmpty=False):
            childrenToMark.append(child)
        elif onlyAtTheEnd:
            # if tag is not managed or not empty, we reinitialize the childrenToMark
            # because we are on a tag that is not considered empty
            childrenToMark = []
    for child in childrenToMark:
        if 'class' in child.attrib:
            child.attrib['class'] = '{0} {1}'.format(markingClass, child.attrib['class'])
        else:
            child.attrib['class'] = markingClass
        # add a title to the tag if necessary
        if tagTitle:
            child.attrib['title'] = tagTitle
    return ''.join([lxml.html.tostring(x,
                                       encoding='utf-8',
                                       pretty_print=pretty_print,
                                       method='html')
                    for x in tree.iterchildren()])


def removeCssClasses(xhtmlContent,
                     css_classes=[],
                     pretty_print=False):
    ''' '''
    tree = _turnToLxmlTree(xhtmlContent)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtmlContent

    for content_tag in CONTENT_TAGS:
        tags = tree.findall('.//{0}'.format(content_tag))
        for tag in tags:
            if 'class' in tag.attrib and tag.attrib['class'].strip() in css_classes:
                del tag.attrib['class']

    return ''.join([lxml.html.tostring(x,
                                       encoding='utf-8',
                                       pretty_print=pretty_print,
                                       method='html')
                    for x in tree.iterchildren()])


def _img_from_src(context, img, portal, portal_url):
    """ """
    # check if it is a local or an external image
    img_src = img.attrib.get('src', None)
    # wrong <img> without src or external image
    if not img_src or (img_src.startswith('http') and not img_src.startswith(portal_url)):
        return
    # here, we have an image contained in the portal
    # either absolute path (http://...) or relative (../images/myimage.png)
    imageObj = None
    # absolute path
    if img_src.startswith(portal_url):
        img_src = img_src.replace(portal_url, '')
        try:
            # get the image but remove leading '/'
            imageObj = portal.unrestrictedTraverse(img_src[1:])
        except (KeyError, AttributeError, NotFound):
            return
    # relative path
    else:
        try:
            imageObj = context.unrestrictedTraverse(img_src)
        # in case we have a wrong resolveuid/unknown_uid, it raises NotFound
        except (KeyError, AttributeError, NotFound):
            return

    # maybe we have a ImageScale instead of the real Image object?
    if isinstance(imageObj, ImageScale):
        imageObj = imageObj.aq_inner.aq_parent
    return imageObj


def _get_image_blob(imageObj):
    """Be defensinve in case this is a wrong <img> with a src
       to someting else than an image... """
    blob = None
    if hasattr(aq_base(imageObj), 'getBlobWrapper') and imageObj.get_size():
        blob = imageObj.getBlobWrapper()
    return blob


def _transform_images(context, xhtmlContent, pretty_print=False, transform_type="path"):
    '''Turn <img> source contained in given p_xhtmlContent to a FileSystem absolute path
       to the .blob binary stored on the server.  This is usefull when generating documents
       with XHTML containing images that are private, LibreOffice is not able to access these
       images using the HTTP request.
       <img src='http://mysite/myfolder/myimage.png' /> becomes
       <img src='/absolute/path/to/blobstorage/myfile.blob'/>,
       external images are left unchanged.
       The image_scale is not kept, so :
       <img src='http://mysite/myfolder/myimage.png/image_preview' /> becomes
       <img src='/absolute/path/to/blobstorage/myfile.blob'/>.'''
    # keep original xhtmlContent in case we need to return it without changes
    originalXhtmlContent = xhtmlContent

    if not originalXhtmlContent or not originalXhtmlContent.strip():
        return originalXhtmlContent

    # if using resolveuid, turn src to real image path
    if 'resolveuid' in xhtmlContent:
        xhtmlContent = ResolveUIDAndCaptionFilter()(xhtmlContent)

    xhtmlContent = "<special_tag>%s</special_tag>" % xhtmlContent
    tree = lxml.html.fromstring(safe_unicode(xhtmlContent))
    imgs = tree.findall('.//img')
    if not imgs:
        return originalXhtmlContent

    portal = api.portal.get()
    portal_url = portal.absolute_url()
    for img in imgs:
        imageObj = _img_from_src(context, img, portal, portal_url)
        if imageObj is None:
            continue

        blob = _get_image_blob(imageObj)
        # change img src only if a blob was found
        blob_path = None
        if blob:
            if transform_type == "path":
                blob_path = blob.blob._p_blob_committed
                if blob_path:
                    img.attrib['src'] = blob_path
            elif transform_type == "data":
                blob_path = blob.blob._p_blob_committed
                if blob_path and blob.content_type.startswith('image/'):
                    img.attrib['src'] = "data:{0};base64,{1}".format(
                        blob.content_type, base64.b64encode(blob.data))

    # use encoding to 'ascii' so HTML entities are translated to something readable
    return ''.join([lxml.html.tostring(x,
                                       encoding='ascii',
                                       pretty_print=pretty_print,
                                       method='html') for x in tree.iterchildren()])


def imagesToPath(context, xhtmlContent, pretty_print=False):
    '''Turn <img> source contained in given p_xhtmlContent to a FileSystem absolute path
       to the .blob binary stored on the server.  This is usefull when generating documents
       with XHTML containing images that are private, LibreOffice is not able to access these
       images using the HTTP request.
       <img src='http://mysite/myfolder/myimage.png' /> becomes
       <img src='/absolute/path/to/blobstorage/myfile.blob'/>,
       external images are left unchanged.
       The image_scale is not kept, so :
       <img src='http://mysite/myfolder/myimage.png/image_preview' /> becomes
       <img src='/absolute/path/to/blobstorage/myfile.blob'/>.'''

    return _transform_images(context, xhtmlContent, pretty_print, transform_type="path")


def imagesToData(context, xhtmlContent, pretty_print=False):
    '''Turn <img> source contained in given p_xhtmlContent to a data:image/png;base64... value.
       External images are left unchanged.
       The image_scale is not kept, so we get the full image.'''

    return _transform_images(context, xhtmlContent, pretty_print, transform_type="data")


def storeImagesLocally(context,
                       xhtmlContent,
                       imagePortalType='Image',
                       store_external_images=True,
                       store_internal_images=True,
                       pretty_print=False,
                       force_resolve_uid=False,
                       replace_not_found_image=True):
    """If images are found in the given p_xhtmlContent,
       we download it and stored it in p_context, this way we ensure that it will
       always be available in case the external/internal image image disappear.
       If p_store_external_images is True, we retrieve external image and store it
       in p_context, if p_store_internal_images is True, we do the same for internal
       images.
       If p_replace_not_found_image is True, an image holding text "Image not available"
       will be used if the image could not be retrieved."""

    def _handle_internal_image(img_src):
        """ """
        # get image from URL
        img_path = img_src.replace(portal_url, '')
        img_path = path.join(portal.absolute_url_path(), img_path)
        # path can not start with a '/'
        img_path = img_path.lstrip('/')

        # right, traverse to image
        try:
            imageObj = portal.unrestrictedTraverse(img_path)
        except (KeyError, AttributeError, NotFound):
            # wrong img_path
            logger.warning(
                'In \'storeImagesLocally\', could not traverse '
                'img_path \'{0}\' for \'{1}\'!'.format(
                    img_path, context.absolute_url()))
            return None, None

        # not an image
        if getattr(imageObj, 'portal_type', None) != 'Image':
            return None, None

        filename = imageObj.getId()
        data = imageObj.getBlobWrapper().data
        return filename, data

    def _handle_external_image(img_src):
        """ """
        # right, we have an external image, download it, stores it in context and update img_src
        try:
            downloaded_img_path, downloaded_img_infos = urllib.urlretrieve(img_src)
        except (IOError, UnicodeError):
            # url not existing
            return None, None

        # not an image
        if not downloaded_img_infos.maintype == 'image':
            return None, None

        # retrieve filename
        filename = 'image'
        disposition = downloaded_img_infos.getheader('Content-Disposition')
        # get real filename from 'Content-Disposition' if available
        if disposition:
            disp_value, disp_params = cgi.parse_header(disposition)
            filename = disp_params.get('filename', 'image')
        # if no 'Content-Disposition', at least try to get correct file extension
        elif hasattr(downloaded_img_infos, 'subtype'):
            filename = '{0}.{1}'.format(filename, downloaded_img_infos.subtype)
        f = open(downloaded_img_path, 'r')
        data = f.read()
        f.close()
        # close and delete temporary file
        if path.exists(downloaded_img_path):
            os.remove(downloaded_img_path)
        return filename, data

    tree = _turnToLxmlTree(xhtmlContent)
    if not isinstance(tree, lxml.html.HtmlElement):
        return xhtmlContent

    imgs = tree.findall('.//img')
    if not imgs:
        return xhtmlContent
    portal = api.portal.get()
    portal_url = portal.absolute_url()
    # make sure context_url ends with a '/' to avoid mismatch if folder holding the
    # image absolute_url starts with context_url
    context_url = context.absolute_url() + '/'
    # adapter to generate a valid id, it needs a container, so it will be context or it's parent
    try:
        name_chooser = INameChooser(context)
    except TypeError:
        parent = context.getParentNode()
        name_chooser = INameChooser(parent)
    # return received xhtmlContent if nothing was changed
    changed = False
    for img in imgs:
        original_img_src = img.get('src', '')

        # we only handle http stored images
        if not original_img_src.startswith('http') and \
           'resolveuid' not in original_img_src and \
           ';base64,' not in original_img_src:
            continue
        filename = data = None
        handled = False
        # external images
        if store_external_images and not original_img_src.startswith(portal_url) and \
           'resolveuid' not in original_img_src:
            filename, data = _handle_external_image(original_img_src)
            handled = True

        # image in portal but not already stored in context
        # handle images using resolveuid
        end_of_url = None
        img_src = original_img_src
        if 'resolveuid' in original_img_src:
            # manage cases like :
            # - resolveuid/2ff7ea3317df43438a09edfa84965b13
            # - resolveuid/2ff7ea3317df43438a09edfa84965b13/image_preview
            # - http://portal_url/resolveuid/2ff7ea3317df43438a09edfa84965b13
            img_uid = original_img_src.split('resolveuid/')[-1].split('/')[0]
            # save end of url if any, like /image_preview
            end_of_url = original_img_src.split(img_uid)[-1]
            if end_of_url == img_uid:
                end_of_url = None
            # if not found, like when copy/paste HTML code containing resolveuid, use '' instead None
            img_src = uuidToURL(img_uid) or ''

        if store_internal_images and \
           img_src.startswith(portal_url) and \
           not img_src.startswith(context_url):
            filename, data = _handle_internal_image(img_src)
            handled = True

        if not filename:
            if handled:
                if replace_not_found_image:
                    filename = "imageNotFound.jpg"
                    f = open(os.path.join(os.path.dirname(__file__), filename), 'r')
                    data = f.read()
                    f.close()
                else:
                    continue
            else:
                continue

        changed = True

        # create image
        name = name_chooser.chooseName(filename, context)
        new_img_id = context.invokeFactory(imagePortalType, id=name, title=name, file=data)
        new_img = getattr(context, new_img_id)
        # store a resolveuid if using it, the absolute_url to image if not
        if force_resolve_uid or \
           original_img_src.startswith('resolveuid') or \
           (HAS_CKEDITOR and
                hasattr(portal.portal_properties, 'ckeditor_properties') and
                portal.portal_properties.ckeditor_properties.allow_link_byuid):
            new_img_src = 'resolveuid/{0}'.format(new_img.UID())
        else:
            new_img_src = new_img.absolute_url()
        if end_of_url:
            new_img_src = '/'.join([new_img_src, end_of_url.strip('/')])

        img.attrib['src'] = new_img_src

    if not changed:
        return xhtmlContent

    return ''.join([lxml.html.tostring(x,
                                       encoding='utf-8',
                                       pretty_print=pretty_print,
                                       method='html') for x in tree.iterchildren()])


def separate_images(xhtmlContent, pretty_print=False):
    """Make sure images are separated in different paragraphs.
       So <p><img .../><img .../></p> is changed to
       <p><img .../></p></p><img .../></p>."""

    tree = _turnToLxmlTree(xhtmlContent)
    # bypass if wrong format or no img
    if not isinstance(tree, lxml.html.HtmlElement) or \
       not tree.xpath('.//img'):
        return xhtmlContent

    # return received xhtmlContent if nothing was changed
    changed = False
    inserted_index = 1
    for elt_index, elt in enumerate(tree.getchildren()):
        if elt.tag not in ('p', 'div'):
            continue
        # only manage <p>/<div> containing several <img>, nothing else
        imgs = elt.xpath('.//img')
        len_imgs = len(imgs)
        if len_imgs > 1:
            # <p> may not contain anything else than <img> or <br>
            contained_tags = [child for child in elt.getchildren()
                              if child.tag not in ('br', )]
            # contained text, if <p> contains <img> and text, we can not separate it
            if len_imgs == len(contained_tags) and not elt.text_content().strip():
                changed = True
                for img_index, img in enumerate(imgs[1:]):
                    new_elt = lxml.html.Element(elt.tag)
                    tree.insert(elt_index + inserted_index + img_index, new_elt)
                    inserted_index += 1
                    # append will move the img, not necessary to remove from elt
                    new_elt.append(img)

    if not changed:
        return xhtmlContent

    return ''.join([lxml.html.tostring(x,
                                       encoding='utf-8',
                                       pretty_print=pretty_print,
                                       method='html') for x in tree.iterchildren()])


def object_link(obj, view='view', attribute='Title', content='', target='', escaped=True):
    """ Returns an html link for the given object """
    href = view and "%s/%s" % (obj.absolute_url(), view) or obj.absolute_url()
    if not content:
        if not hasattr(obj, attribute):
            attribute = 'Title'
        content = getattr(obj, attribute)
        if callable(content):
            content = content()
    if target:
        target = ' target="{}"'.format(target)
    if escaped:
        content = escape(content)
    return u'<a href="%s"%s>%s</a>' % (href, target, safe_unicode(content))


def is_html(text):
    """Check that given p_text is HTML."""
    text = text or ''
    mtr = api.portal.get_tool('mimetypes_registry')
    # taken from Products.Archetypes/Field.py
    try:
        d, f, mimetype = mtr(text[:8096], mimetype=None)
    except UnicodeDecodeError:
        d, f, mimetype = mtr(len(text) < 8096 and text or '', mimetype=None)
    return str(mimetype).split(';')[0].strip() == 'text/html'


def unescape_html(html):
    """Unescape an html text, turn HTML entities to encoded entities."""
    if not html:
        return html
    is_unicode = isinstance(html, unicode)
    parser = HTMLParser()
    html = parser.unescape(html)
    return html if is_unicode else html.encode('utf-8')



# -*- coding: utf-8 -*-
"""Installer for the imio.helpers package."""

from setuptools import find_packages
from setuptools import setup


long_description = (
    open('README.rst').read() + '\n' + open('CHANGES.rst').read() + '\n')

setup(
    name='imio.helpers',
    version='0.80',
    description="Various helper methods for development.",
    long_description=long_description,
    # Get more from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 4.3",
        "Framework :: Plone :: 5.2",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords='plone helpers utils dev imio',
    author='Simon Delcourt',
    author_email='simon.delcourt@imio.be',
    url='https://github.com/imio/imio.helpers',
    download_url='https://pypi.org/project/imio.helpers',
    license='GPL',
    packages=find_packages('src', exclude=['ez_setup']),
    namespace_packages=['imio'],
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'collective.fingerpointing',
        'collective.monkeypatcher',
        'future>=0.18.2',
        'imio.pyutils>=0.31',
        'natsort<7',
        'plone.api>1.9.1',
        'plone.app.intid',
        'plone.app.relationfield',
        'plone.dexterity',
        'setuptools',
        'six',
        'Plone',
        'z3c.unconfigure',
    ],
    extras_require={
        'test': [
            'collective.behavior.talcondition',
            'collective.MockMailHost',
            'ipdb',
            'plone.app.dexterity',
            'plone.app.testing',
        ],
        'pdf': [
            'PyPDF2',
            'reportlab',
        ],
        'lxml': [
            'cssselect<1.2; python_version == "2.7"',
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
